print("import: 'autograd'")
import autograd

print("import: 'autograd.numpy'")
import autograd.numpy

print("import: 'autograd.scipy'")
import autograd.scipy

print("import: 'autograd.scipy.stats'")
import autograd.scipy.stats

print("import: 'autograd.misc'")
import autograd.misc

